package org.w3c.rdf.util.xml;

import org.w3c.rdf.model.*;
import org.w3c.rdf.implementation.model.StatementImpl;
import org.w3c.rdf.syntax.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;

import java.util.*;
import java.io.*;
import java.net.*;

public class GenericParser extends DefaultHandler implements RDFParser { /*EntityResolver, DTDHandler, DocumentHandler, */

  static final public String REVISION = "Generic RDF/XML parser v0.2 2000-09-12";

  protected ErrorHandler errorHandler = null;
  protected Locator locator = new LocatorImpl();
  protected InputSource source;
  protected Element current;

  protected NodeFactory nodeFactory;
  protected RDFConsumer consumer;

  // collect all namespaces here
  protected Hashtable namespaces = new Hashtable();

  public GenericParser () {

    this(true);
  }

  public GenericParser (boolean warn) {

    initXMLParser(warn);
  }

  protected void initXMLParser(boolean warn) {

    if(System.getProperty("org.xml.sax.parser") == null) {
      if(warn)
	System.err.println("Warning: using the default XML parser AElfred (com.microstar.xml.SAXDriver).\n" +
			   "Override for IBM xml4j is: java -Dorg.xml.sax.parser=com.ibm.xml.parser.SAXDriver org.w3c.rdf.SiRPAC");
      try {
	System.getProperties().put("org.xml.sax.parser", "org.brownell.xml.aelfred2.SAXDriver");
      } catch (Exception any) {
	if(warn)
	  System.err.println("Warning: could not set default parser");
      }
    }
  }

  /**
   * Generate a warning message as a string
   */
    public void addWarning (String sMsg) throws SAXException {
      errorHandler.warning(new SAXParseException(sMsg, locator));
    }
  
  /**
   * Generate an error message as a string
   */
   public void addError (String sMsg) throws SAXException {
     errorHandler.error(new SAXParseException(sMsg, locator));
   }
  
//   public void setDocumentLocator (Locator locator) {

//     this.locator = locator;
//   }
  
  public void setErrorHandler (ErrorHandler handler) {
    this.errorHandler = handler;
  }

  protected String getSourceURI() {

    return source.getSystemId();
  }

  public void startElement (String uri, String name, String qName, Attributes atts) throws SAXException {

    _startElement(uri, name, atts);
    System.out.println("Start element: " + current.getName());
  }

  public Enumeration getNamespaces() {

    return namespaces.keys();
  }

  protected void addNamespace(String uri) {

    if(!"".equals(uri))
      namespaces.put(uri, uri);
  }

  protected void _startElement (String uri, String name, Attributes al) throws SAXException {

    Element el = createElement();
    el.setName(uri + name);
    addNamespace(uri);

    // add attributes
    for (int x = 0; x < al.getLength(); x++) {
      String aURI = al.getURI(x);
      addNamespace(aURI);
      // take element's namespace if none specified
      if("".equals(aURI))
	aURI = uri;
      String aName = aURI + al.getLocalName(x);
      String aValue = al.getValue (x);
      //      System.out.println("Adding property: " + aName + "=" + aValue);
      el.setAttribute(aName, aValue);
    }

    if(current == null)
      current = el;
    else {
      current.setChild(el);
      el.setParent(current);
      current = el;
    }

  }

  protected Element createElement() {
    return new Element();
  }

  public void endElement (String uri, String name, String qName) throws SAXException {

    System.out.println("End element: " + current.getName());
    _endElement(uri, name);
  }

  protected void _endElement (String uri, String name) throws SAXException {

    current = current.getParent();
  }

  protected boolean preserveWhiteSpace() {
    return false;
  }

  public void characters (char ch[], int start, int length)
    throws SAXException {
  
      String s = new String (ch, start, length);
      String sTrimmed = s.trim();
      if (sTrimmed.length() > 0 || preserveWhiteSpace()) {
	// System.out.println("Adding characters: \"" + s + "\"");
	current.setValue(sTrimmed);
      }
  }

  protected Statement createStatement(Resource subject, Resource predicate, RDFNode object) throws ModelException {

    Statement st;
    consumer.addStatement(st = nodeFactory.createStatement(subject, predicate, object));
    return st;
  }

  public static XMLReader createParser (String className) {

    try {
      // Get the named class.
      Class c = Class.forName(className);
      // Instantiate the parser.
      return (XMLReader)(c.newInstance());
    } catch (Exception any) {
      throw new RuntimeException("Could not instantiate XML parser: " + any);
    }
  }    

  public void parse(InputSource source, RDFConsumer consumer) throws SAXException {

    this.source = source;
    this.consumer = consumer;

    try {
      this.nodeFactory = consumer.getNodeFactory();
      consumer.startModel();

      XMLReader p = XMLReaderFactory.createXMLReader();
      //      Parser p = ParserFactory.makeParser();

      // Register the handlers
      //      p.setEntityResolver(this);
      //      p.setDTDHandler (this);
      p.setContentHandler(this);
      p.setErrorHandler (errorHandler != null ? errorHandler : this);

      p.parse ( source );

      consumer.endModel();

    } catch (Exception any) {
      if(any instanceof SAXException)
      	throw (SAXException)any;
      else
      	throw new SAXException("Fatal error", any);
    }
  }

  public static InputSource getInputSource(String urlStr) throws MalformedURLException, IOException {

    InputStream in = null;
    URL url = null;

    if(urlStr != null) {

      try {
	url = new URL (urlStr);
      } catch (Exception e) {
	url = new URL ("file", null, urlStr);
      }
      in = url.openStream();
      
    } else {
      // read from stdin
      in = System.in;
    }
    
    Reader r = new BufferedReader(new InputStreamReader(in));
    InputSource source = new InputSource(r);
    if(urlStr != null)
      source.setSystemId(url.toString());

    return source;
  }

  protected static void _main(String url, RDFParser parser) throws IOException, MalformedURLException {

    InputSource source = getInputSource(url);
    RDFConsumer consumer = new DumpConsumer();
    ErrorStore handler = new ErrorStore();

    parser.setErrorHandler(handler);

    try {
      parser.parse(source, consumer);
    } catch (SAXException se) {
      System.err.println("Error during parsing: " + se.getMessage());
      Exception ex = (se.getException() != null) ? se.getException() : se;
      ex.printStackTrace(System.err);
    }
    
    String sErrors = handler.errors ();
    if (sErrors != null && sErrors.length() > 0)
      System.err.println ("Errors during parsing:\n"+sErrors);
  }

  public static void main (String args[]) throws Exception {

    if(args.length != 1) {
      System.err.println("Usage: java -Dorg.xml.sax.parser=<classname> org.w3c.rdf.syntax.xml.GenericParser " +
			 //			 "[-fetch_schemas | -f] [-stream | -s] [-robust | -r] [ URI | filename ]");
			 "[ URI | filename ]");
      System.err.println ("This is revision " + REVISION);
      System.exit(1);
    }

    GenericParser gp;
    _main(args[0], gp = new GenericParser());

    Enumeration en = gp.getNamespaces();
    while(en.hasMoreElements())
      System.out.println("Namespaces used: " + en.nextElement());
  }

}
